import sys, ZODB, ZODB.FileStorage
from BTrees.OOBTree import OOBTree
from persistent import Persistent
from persistent.list import PersistentList
from persistent.mapping import PersistentMapping

import threading, tas, os, os.path, time, random
import sets, sys

import config, logcheck



COUNT = 0
class Transaction(tas.Transaction):
    def __init__(_, ctx, db):
        tas.Transaction.__init__(_, ctx)
        _.__db = db

    def _commit(_):
        get_transaction().commit()

    def _rollback(_):
        get_transaction().abort()

    def _create_conn(_):
        _.__conn = _.__db.open()

    def _release_conn(_):
        _.__conn.close()
        del _.__conn

    def __get_connection(_):
        return _.__conn

    connection = property(__get_connection)
    
    
class Context(tas.AppContext):
    def __init__(_, db):
        tas.AppContext.__init__(_)
        _.db = db

    def _create_transaction(_):
        return Transaction(_, _.db)


class ThreadProp(object):
    __metaclass__ = tas.TransactionAware
    """
    A thread prop is creates a property that depends on a data obcject local to
    the current thread.
    """
    def __init__(_, name):
        _.prop_name = name

    def get_set(_):
        return [_.get, _.set]

    def get(_, parent):
        """
        tas::active
        """
        return getattr(parent._data, _.prop_name)

    def set(_, parent, value):
        """
        tas::active
        """        
        setattr(parent._data, _.prop_name, value)


class ChildData(Persistent):
    def __init__(_, payload):
        _.payload = payload

class ConflictPersistentList(PersistentList):
    def _p_resolveConflict(_, oldState, savedState, newState):
        print "You shouldn't see this..."
        print oldState, newState, savedState
        return oldState
        
class ParentData(Persistent):
        
    def __init__(_, name):
        
        _.name = name
        _.items = ConflictPersistentList()

        
class Child(object):
    __metaclass__ = tas.TransactionAware
    
    def __init__(_, parent, key):
        _.__parent = parent
        _.__key = key
        
    def __get_data(_):
        data = _.__parent.items[_.__key]
        return data

    _data = property(__get_data)

    payload = property(*ThreadProp("payload").get_set())

class Parent(object):
    __metaclass__ = tas.TransactionAware

    def __init__(_, name):
        _.name = name
        dd = _._data
        _._item_map = {}


    items = property(*ThreadProp("items").get_set())

    def len(_):
        """
        returns the 
        tas::active
        """
        l = len(_.items)
        d = l - len(_._item_map.keys())
        return l, d
    
    def __get_data(_):
        """
        tas::active
        """
        conn = tas.get_context().transaction.connection
        return conn.root()['parents'][_.name]

    _data = property(__get_data)

    def add_child(_):
        global COUNT
        COUNT += 1
        child = ChildData("")
        _.items.append(child)
        index = len(_.items) - 1
        child.payload = index
        _._item_map[index] = Child(_, index)
        #print _.items, _._item_map

    def all_childs(_):
        print _.items
        return _.items



def create_db(name="conflict.db"):
    if os.path.exists(name):
        os.remove(name)
    db = ZODB.DB(ZODB.FileStorage.FileStorage(name))
    return db

def setup_parent(db, name="father"):
    ctx = tas.get_context()
    t = ctx.create_transaction()
    t.begin()    
    r = t.connection.root()
    parents = {name : ParentData(name)}
    r['parents'] = parents
    parent = Parent(name)
    t.commit()
    return parent

def create_context(*args):
    return Context.create_context(*args)

    

class Application(threading.Thread):
    __metaclass__ = tas.TransactionAware

    def __init__(_, context, parent, name="App"):
        threading.Thread.__init__(_)
        _.setDaemon(True)
        _.name = name
        _.context = context
        _.parent = parent
        _.ops = [_.add, _.length]
        _.running = True
        _.stopped = False        
        _.start()
        
    def run(_):
        while _.running:
            _.test()
            time.sleep(.5)
        _.stopped = True

    def stop(_):
        _.log("stopping")        
        _.running = False
        while not _.stopped:
            pass
        _.log("stopped")
        return _
    
    def test(_):
        """
        tas::sync, create, active
        """
        _.ops[random.randint(0, len(_.ops) - 1)](_.parent)

    def add(_, p):
        """
        tas::autocommit
        """
        #print tas.get_context().transaction
        p.add_child()
        time.sleep(2)        
        p.add_child()
        time.sleep(2)        
        _.log("added childs")

        
    def length(_, p):
        _.log("len: %s" % repr(p.len()))


    def log(_, msg):
        print "%s: %s" % (_.name, msg)
        

def main():
    conf, cn = config.process_config()
    logcheck.setup_logging(conf)
    
    db = create_db()
    rootctx = create_context(db)
    rootctx.bind()
    parent = setup_parent(db)
    rootctx.unbind()
    apps = []
    for i in xrange(3):
        ctx = create_context(db)
        app = Application(ctx, parent, name="App_%i" % i)
        apps.append(app)
    time.sleep(20)
    [app.stop() for app in apps]
    rootctx.bind()
    rootctx.create_transaction()
    for i, child in enumerate(parent.all_childs()):
        print i, child.payload
    rootctx.unbind()

    print COUNT
if __name__ == "__main__":
    main()
    
