"""
tas - transaction services
"""

import threadlocal, UserDict, logging, re, threading, traceback, sys
import StringIO

class AppContext(UserDict.UserDict, object):
    CONTEXTS = []

    def __init__(_):
        UserDict.UserDict.__init__(_)
        _.__pre_hooks = []
        _.__post_hooks = []
        _.__lock = threading.RLock()
        _.__level = 0
        _.logger = logging.getLogger("tas.context")
        
        
    def bind(_):
        _.__level += 1        
        if _.__level > 1:
            return
        _.logger.debug("binding context %s" % id(_))
        tc = threadlocal.ThreadedContext()
        if tc.has_key("__app_context__") and not tc["__app_context__"] is _:
            _.logger.warning("rebinding context!")
            
        tc["__app_context__"] = _

    def unbind(_):
        _.__level -= 1        
        _.logger.debug("bind level: %i in context %i" % (_.__level, id(_)))
        if _.__level > 0:
            return
        _.logger.debug("unbinding context")
        tc = threadlocal.ThreadedContext()
        del tc["__app_context__"]
        if _.has_key("__transaction__"):
            t = _["__transaction__"]
            t.commit_logic()
    
    def acquire(_):
        _.__lock.acquire()


    def release(_):

        _.__lock.release()

        
        
    def create_transaction(_):
        if not _.has_key("__transaction__") or not _["__transaction__"].alive:
            _["__transaction__"] = _._create_transaction()
        return _["__transaction__"]

    def __get_transaction(_):
        return _["__transaction__"]

    transaction = property(__get_transaction)
    
    def get_context():
        return threadlocal.ThreadedContext()["__app_context__"]

    get_context = staticmethod(get_context)


    def create_context(cls, *args):
        c = cls(*args)
        cls.CONTEXTS.append(c)
        return c

    create_context = classmethod(create_context)
    
    def receive(_, msg):
        sender, payload = msg
        
        if not sender is _ and callable(payload):
            payload(_)
            
    def emit(_, payload):
        _.__broadcast((_, payload))

    def __broadcast(cls, msg):
        for ctx in cls.CONTEXTS:
            ctx.receive(msg)

    __broadcast = classmethod(__broadcast)
    
    def pre_hooks(_):
        _.__pre_hooks = [hook for hook in _.__pre_hooks if hook.process(_)]

    def post_hooks(_):
        _.__post_hooks = [hook for hook in _.__post_hooks if hook.process(_)]

    def add_pre_hook(_, hook):
        _.__pre_hooks.append(hook)

    def add_post_hook(_, hook):
        _.__post_hooks.append(hook)


    def _create_transaction(_):
        raise "not implemented"

def get_context():
    return AppContext.get_context()


        
class Transaction(object):
    """
    Transaction has two states:

     - alive: This flag indicates if the transactions livecycle has ended or not.
     The Transaciton is alive when its created and becomes dead when either rollback or commit are
     called.
     - active: The Transaction is active between a begin and either commit or rollback
    
    """
    def __init__(_, ac):
        _.ac = ac
        _.__alive_flag = True
        _.__active_flag = False
        _.__commit_flag = False
        _.__rollback_flag = False
        _.logger = logging.getLogger("tas.transaction")        

    def begin(_):
        _.__alive()
        _.__active_flag = True
        _.logger.debug("begin")
        _._create_conn()

    def commit(_):
        _.__alive()
        _.logger.debug("commit")
        _.__commit_flag = True

        
    def rollback(_):
        _.__alive()        
        _.__rollback_flag = True
        _.logger.debug("rollback")

    def commit_logic(_):
        _.logger.debug("commit_logic")
        _.__alive(True)
        _.__active_flag = False
        _.logger.debug("c: %s, r: %s" % (_.__commit_flag,  _.__rollback_flag))
        try:
            if not _.__commit_flag  and not _.__rollback_flag:
                return
            elif _.__rollback_flag:
                _._rollback()
            else:
                _._commit()
        finally:
            _._release_conn()

    def __alive(_, kill_me=False):
        #print threading.currentThread()        
        if not _.alive:
            _.logger.error("Transaction is dead!")
            raise "Transaction is dead!"
        if kill_me:
            #traceback.print_stack()
            _.logger.debug("killing transaction")
            _.__alive_flag = False

    def __is_alive(_):
        return _.__alive_flag
    
    alive = property(__is_alive)

    def __is_active(_):
        return _.__active_flag
    
    active = property(__is_active)
    
    
    def _commit(_):
        raise "not implemented"

    def _rollback(_):
        raise "not implemented"

    def _create_conn(_):
        raise "not implemented"

    def _release_conn(_):
        raise "not implemented"

    def __repr__(_):
        return "Active: %s, Alive: %s" % (_.active, _.alive)


def w_create(name):
    """
    Ensures that there is a transaction in the apps context.
    """    
    def _w_create(_, *args, **kwargs):
        """

        """
        ctx = get_context()
        try:
            t = ctx.transaction
            if not t.alive:
                ctx.create_transaction()
        except KeyError:
            t = ctx.create_transaction()
        res = getattr(_, name)(*args, **kwargs)
        return res

    return _w_create

def w_active(name):
    """
    Ensures that the transaction is active.
    """    
    def _w_active(_, *args, **kwargs):
        t = get_context().transaction
        if not t.active:
            t.begin()
        res = getattr(_, name)(*args, **kwargs)
        return res

    return _w_active

def w_autocommit(name):
    """
    Will issue a commit so that the unbinding of the context will commit automatically. If an exception
    occured, the transaction is rolled back.
    """
    def _w_autocommit(_, *args, **kwargs):
        """

        """
        t = get_context().transaction
        try:
            res = getattr(_, name)(*args, **kwargs)
            t.commit()
            return res
        except:
            t.rollback()
            raise

    return _w_autocommit

def w_bind(name):
    def _w_bind(_, *args, **kwargs):
        """

        """
        # bind the context        
        _.context.bind()
        try:
            res = getattr(_, name)(*args, **kwargs)
            return res
        finally:
            _.context.unbind()
            

    return _w_bind

def w_sync(name):
    """
    binds the app context and syncs on it.
    """
    def _w_sync(_, *args, **kwargs):
        # sync the context        
        _.context.acquire()
        _.context.bind()
        try:
            try:
                res_fetched = False                
                _.context.logger.debug("call: %s" % name)
                res = getattr(_, name)(*args, **kwargs)
                res_fetched = True
                _.context.unbind()
                return res
            except:
                # if the res hasn't been fetched then we got an application
                # exception that caused the unbind above to be ignored, so we
                # need to unbind.
                if not res_fetched:
                    _.context.unbind()
                raise
        finally:
            _.context.release()
    return _w_sync

    
class TransactionAware(type):
    TAS_REX = re.compile("tas::([^, ]+(, *[^, ]+)*)")


    WRAPPERS = {
        "create" : w_create,
        "active" : w_active,
        "bind" : w_bind,
        "sync" : w_sync,
        "autocommit" : w_autocommit,                
        }
    def __new__(_, name, bases, dict):

        for k,v in [(k, v) for k,v in dict.items() if callable(v)]:
            _.wrap(k,v,_.get_directives(v), dict)

        return super(TransactionAware, _).__new__(_, name, bases, dict)

    def wrap(_, name, fun, ds, dict, level=0):
        if ds:
            d, rest = ds[0], ds[1:]
            key = "_taw_%s_%i" % (name, level)
            try:
                w_fun = _.WRAPPERS[d](key)
                dict[key] = fun
                _.wrap(name, w_fun, rest, dict, level+1)
            except KeyError, e:
                print "No transaction aware property named %s" % d
                raise e
        else:
            
            dict[name] = fun

    wrap = classmethod(wrap)
    
    def get_directives(_, v):
        try:
            doc = v.__doc__
            res = []
            if doc:
                for l in doc.split("\n"):
                    m = _.TAS_REX.match(l.strip())
                    if m:
                        res += [s.strip() for s in m.group(1 ).split(",")]
            res.reverse()
            return res
        except KeyError:
            return []
    get_directives = classmethod(get_directives)
    
if __name__ == "__main__":
    class TestT(Transaction):
        def _commit(_):
            print "@commit"

        def _rollback(_):
            print "@rollback"

        def _get_conn(_):
            print "@get_conn"

        def _release_conn(_):
            print "@release_conn"


    class TestC(AppContext):
        def _create_transaction(_):
            return TestT(_)

    class Foo(object):
        __metaclass__ = TransactionAware

        def bar(_):
            """
            tas::create, active
            """
            print "bar"
            
    
    ctx = TestC.create_context()
    f = Foo()
    f.bar()

    print ctx.__class__
    t = ctx.transaction
    print t.alive
    print t.active
    t.begin()
    print t.alive
    print t.active
    t.commit()
    print t.alive
    print t.active
    try:
        
        t.begin()
    except:
        pass


    
    
